/*
 * "Copyright (c) 2006 University of Southern California. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without written
 * agreement is hereby granted, provided that the above copyright
 * notice, the following two paragraphs and the author appear in all
 * copies of this software.
 *
 * IN NO EVENT SHALL UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND UNIVERSITY OF SOUTHERN CALIFORNIA
 * HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS."
 */

/**
 *This class is the control interface for Querying
 *Author: @Avinash Sridharan, asridhar@usc.edu
 		  @Sundeep Pattem
 *		@Joon Ahn <joonahn@usc.edu>
 */

 import net.tinyos.message.*;
 import net.tinyos.packet.*;
 import net.tinyos.util.*;
 import java.util.Timer;
 import java.util.TimerTask;
 import java.io.*;

 public class BaseController implements MessageListener{

	//App commands
	final static byte CMD_GET_FLASHDATA = 1;

	//Flash commands
	final static byte JA_LFC_NOP			= 0;
	final static byte JA_LFC_GETALL 		= 1;
	final static byte JA_LFC_LOG 		= 2;
	final static byte JA_LFC_LOGDONE	= 3;
	final static byte JA_LFC_ACK			= 4;
	final static byte JA_LFC_BUSY		= 5;
	final static byte JA_LFC_INVALID		= 6;

	// Flood Type Of packet
	final static short START_FLOOD = 24;
	final static short FORWARD_FLOOD = 25;

	// Flood Operations
	final static byte FLOOD_OP_NOP = 0;			// no operation
	final static byte FLOOD_OP_SYNC_LOG = 1;
	final static byte FLOOD_OP_ERASE_LOG = 2;
	final static byte FLOOD_OP_BEGIN_EXPERIMENT = 3;	// not yet ready. don't use this	// only applied to START_FLOOD and "src" will be experiment number.


	// ID commands
	final static byte JA_ID_OP_GET		= 1;
	final static byte JA_ID_OP_DATA		= 2;

	final static short TOS_BCAST_ADDR = (short)0xffff;

	private MoteIF moteIF;
	static MoteIF mote;

	static short 	dest_node = TOS_BCAST_ADDR;
	static short	src_node = 1;
	static short	num_floods = 1;
	static short	op_code = 0;
	static short	m_ttl = 254;
	static String command = "";

	static BaseController baseController;

	//file writing variables
//	static FileWriter fw;
//	static BufferedWriter bw;

	// JA
	volatile boolean m_isLogDone;
	volatile boolean m_isIdRecved;
	short m_lastACK;

	public BaseController(MoteIF moteIF){
		this.moteIF = moteIF;
		//register to the serial forwarder packet.
		this.moteIF.registerListener( new JaLogFloodMsg(), this );
		this.moteIF.registerListener( new JaIDMsg(), this );

		m_isLogDone = false;
		m_isIdRecved = false;
		m_lastACK = -1;
	}

	public static void usage(){
		System.out.println("java BaseController -c <command> [-d dest node] [-p power] [-s src_id] [-i flood_id] [-n num_of_floods] [-o operation] [-t ttl] [-f serial_forwarder_address]");
		System.out.println("<command>:");
		System.out.println("\t\tGET_FLASHDATA:  Get the data in the flash memory." );
		System.out.println("\t\tSEND_GETID:  " );
		System.out.println("\t\tSEND_FLOODING:  " );
		System.out.println( "\t\tSTART_EXP <-i staring_flood_id> <-n num_of_floods>" );
//		System.out.println("\t\tSYS_TIMER_START: Start the system timer.");
//		System.out.println("\t\tSYS_SET_POWER: Set the power level.");
//		System.out.println("\t\tSYS_STOP_LOGGING: Erasing the flash memory.");
//		System.out.println("\t\tSYS_TEST_NODE_ID: Erasing the flash memory.");
//		System.out.println("\t\tSTART_RED_LED: to turn Red LED on.");
//		System.out.println("\t\tPRR_START : to start the PRR timer.");
//		System.out.println("\t\tPRR_SEND_TO_BS: Send PRR table to the");
//		System.out.println("\t\t				Base station for a specific Node");
//		System.out.println("\t\tRW_START: to start the random walk (Use -q1/q2 to start a query and get a reply back).");
//		System.out.println("\t\tRW_STOP: to stop the random walk.");
//		System.out.println("\t\tREAD_FLASH: to read logged data from flash.");
//		System.out.println("\t\tERASE_FLASH: Erasing the flash memory.");
//		System.out.println("\t\tREAD_LINE: to read a particular line from the flash.");
	}

	public synchronized void messageReceived(int to, Message message)
 	{try {
//		if ( message.getClass().getName().equals( "JaLogFloodMsg" ) )
		if ( message.amType() == JaLogFloodMsg.AM_TYPE )
		{
			JaLogFloodMsg msgIn = (JaLogFloodMsg) message;
			InputStream is = new ByteArrayInputStream( message.dataGet(), msgIn.offset_data( 0 ), message.dataGet().length - msgIn.offset_data( 0 ) );
			byte[] bufLogData = new byte[ message.dataGet().length - msgIn.offset_data( 0 ) ];

			try
			{
//				System.out.println( "offset: " + msgIn.offset_data( 0 ) + "\tlenght: " + message.dataGet().length + "\tbufLength: " + bufLogData.length );
				is.read( bufLogData, 0, bufLogData.length );
			}
			catch ( IOException ex )
			{
				System.err.println( "Unable to read from ByteArrayInputStream." );
				return;
			}

			JaLogData logdata = new JaLogData( bufLogData );

			if ( command.equals( "GET_FLASHDATA" ) )
			{
				// send an appropriate ACK
				if ( msgIn.get_op() == JA_LFC_LOG )
				{
					if ( m_lastACK != msgIn.get_seqno() )
					{
						m_lastACK = msgIn.get_seqno();

						System.out.println( msgIn.get_src()
										+ "\t" + logdata.get_time()
										+ "\t" + logdata.get_id()
										+ "\t" + logdata.get_src()
										+ "\t" + logdata.get_dest()
										+ "\t" + logdata.get_top()
										+ "\t" + logdata.get_op()
										+ "\t" + logdata.get_power()
										+ "\t" + logdata.get_commType()
										+ "\t" + logdata.get_ttl() );
					}

//					Thread.sleep( 150 );

					JaLogFloodMsg msgout = new JaLogFloodMsg();
					msgout.amTypeSet( msgout.AM_TYPE );
					msgout.set_src( src_node );
					msgout.set_op( JA_LFC_ACK );
					msgout.set_size( (short) msgout.DEFAULT_MESSAGE_SIZE );
					msgout.set_seqno( msgIn.get_seqno() );

					try { moteIF.send( msgIn.get_src(), msgout ); }
					catch ( IOException ex )
					{
						System.err.println( "Unable to send the logflood ACK packet." );
					}
				}
				else if ( msgIn.get_op() == JA_LFC_LOGDONE )
				{
					if ( m_lastACK != msgIn.get_seqno() )
					{
						m_lastACK = msgIn.get_seqno();
						System.out.println( msgIn.get_src() + "\t" + "Log Done." );
						System.out.println( "\twith src=" + msgIn.get_src() + "\top=" + msgIn.get_op() + "\tsize=" + msgIn.get_size()
										+ "\tseqno=" + msgIn.get_seqno() );
					}

					JaLogFloodMsg msgout = new JaLogFloodMsg();
					msgout.amTypeSet( msgout.AM_TYPE );
					msgout.set_src( src_node );
					msgout.set_op( JA_LFC_ACK );
					msgout.set_size( (short) msgout.DEFAULT_MESSAGE_SIZE );
					msgout.set_seqno( msgIn.get_seqno() );

					try { moteIF.send( msgIn.get_src(), msgout ); }
					catch ( IOException ex )
					{
						System.err.println( "Unable to send the logflood ACK packet." );
					}

					this.m_isLogDone = true;
//					System.exit( 0 );
				}
				// send the ACK
			}
		}
		else if ( message.amType() == JaIDMsg.AM_TYPE )
		{
			JaIDMsg inmsg = (JaIDMsg) message;

			if ( inmsg.get_op() == JA_ID_OP_DATA )
			{
				System.out.println( "== ID received ==\n" +
					                      "src:\t\t" + inmsg.get_src() + "\n" +
					                      "app name:\t" + new String( inmsg.get_name() ) + "\n" +
					                      "version:\t" + inmsg.get_majorVersion() + "." + inmsg.get_minorVersion() );

				m_isIdRecved = true;
			}
		}
	}
	catch ( Exception ex )
	{
		/* DO NOTHING */
		System.err.println( "Exception occurred." );
	}	}

	public static void main(String[] args) throws Exception{
		String source = "sf@localhost:9001";
//		String source = "serial@/dev/ttyUSB1:telosb";
		short power = 31;
		short final_dest = 0;
		short pkt_id = 1;
//		short rw_start_node = 0;
//		short query_type = 0;
//		short query_param = 255;
//		short rw_mode = 0;
//		short rw_id = 0;
//		int num_nodes;

		if (args.length < 3){
			BaseController.usage();
			System.exit(1);
		}

		for ( int i = 0; i < args.length - 1 ; i++){
			if (args[i].equals("-c")){
				i++;
				//this is the command.
				command = args[i];
			} else if (args[i].equals("-p")){
				i++;
				//this is the power that needs to be set in the motes.
				//default is 31.
				power = (short)Integer.parseInt(args[i]);
			} else if (args[i].equals("-d")){
				i++;
				//this is the destination node to which the packet is directed
				dest_node =(short)Integer.parseInt(args[i]);
			}
			else if (args[i].equals("-s")){
				i++;
				//this is the destination node to which the packet is directed
				src_node =(short)Integer.parseInt(args[i]);
			}
			else if ( args[i].equals("-i"))
			{
				i++;
				pkt_id = (short)Integer.parseInt( args[i] );
			}
			else if ( args[i].equals( "-n" ) )
			{
				i++;
				num_floods = (short)Integer.parseInt( args[i] );
			}
			else if ( args[i].equals( "-o" ) )
			{
				i++;
				op_code = (short)Integer.parseInt( args[i] );
			}
			else if ( args[i].equals( "-t" ) )
			{
				i++;
				m_ttl = (short)Integer.parseInt( args[i] );
			}
			else if ( args[i].equals( "-f" ) )
			{
				i++;
				source = args[i];
			}
			else
			{
				System.err.println( "warning: unknown parameter - " + args[i] );
			}
		}

//		num_nodes = Integer.parseInt(args[args.length - 1]);
		if (command.equals("")){
			System.out.println("ERROR: Please specify a command to be"
			+"executed.");
			BaseController.usage();
			System.exit(1);
		}

		System.out.println("Creating the mote IF interface");
		//Build the Phoenix source
		PhoenixSource phoenix;

		if (source == null){
			phoenix = BuildSource.makePhoenix(PrintStreamMessenger.err);
		} else {
			phoenix = BuildSource.makePhoenix(source, PrintStreamMessenger.err);
		}

		mote = new MoteIF(phoenix);

		baseController = new BaseController(mote);
//		RWMsg RWPkt = new RWMsg();
//		PRRMsg PRRPkt = new PRRMsg();
//		RWFlashMsg RWFlashPkt = new RWFlashMsg();
//		EmpSysMsg EmpSysPkt = new EmpSysMsg();

		System.out.println("System parameters: ");
		System.out.println("\tCommand: "+command);
		System.out.println("\tPower: "+power);
		System.out.println( "\tDest. node: " + dest_node );


		if (command.equals( "GET_FLASHDATA" ))
		{
			JaLogFloodMsg mymsg = new JaLogFloodMsg();
			mymsg.set_src( src_node );
			mymsg.set_op( JA_LFC_GETALL );
			mymsg.set_size( (short)mymsg.DEFAULT_MESSAGE_SIZE );
			mymsg.set_seqno( (short) 0 );
			mymsg.amTypeSet( mymsg.AM_TYPE );

			try
			{
				mote.send( dest_node, mymsg );
				System.out.println( "Message sent." );
			}
			catch ( IOException ex )
			{
				System.err.println( "Unable to send the logflood packet." );
			}

			final_dest = dest_node;

			while ( !baseController.m_isLogDone )
			{
				Thread.sleep( 10 );
			}

			System.exit( 0 );
		}
		else if ( command.equals( "SEND_FLOODING" ) )
		{
			for ( short i = 0 ; i < num_floods ; ++i )
			{
				FloodMsg mymsg = new FloodMsg();
				mymsg.set_id( pkt_id + i );
				mymsg.set_src( src_node );
				mymsg.set_dest( dest_node );
				mymsg.set_top( (short)25 );					// FORWARD_FLOOD
				mymsg.set_op( op_code );			// FLOOD_OP_NOP
				mymsg.set_ttl( m_ttl );
				mymsg.set_power( power );

				try
				{
					mote.send( dest_node, mymsg );
					System.out.println( "Message sent." );
				}
				catch ( IOException ex )
				{
					System.err.println( "Unable to send the logflood packet." );
				}

				Thread.sleep( 1000 );			// 1000 millisec
			}

			System.exit( 0 );

		}
		else if ( command.equals( "SEND_GETID" ) )
		{
			JaIDMsg mymsg = new JaIDMsg();
			mymsg.set_src( src_node );
			mymsg.set_op( JA_ID_OP_GET );

			mote.send( dest_node, mymsg );
			System.out.println( "Message sent" );

			while ( !baseController.m_isIdRecved )
			{
				Thread.sleep( 10 );
			}

			System.exit( 0 );
		}
		else if ( command.equals( "START_EXP" ) )
		{
			FloodMsg mymsg = new FloodMsg();
			mymsg.set_id( pkt_id );
			mymsg.set_src( num_floods );
			mymsg.set_dest( dest_node );
			mymsg.set_top( START_FLOOD );
			mymsg.set_op( FLOOD_OP_BEGIN_EXPERIMENT );
			mymsg.set_ttl( m_ttl );
			mymsg.set_power( power );

			try
			{
				mote.send( dest_node, mymsg );
				System.out.println( "Message sent." );
			}
			catch ( IOException ex )
			{
				System.err.println( "Unable to send the logflood packet." );
			}

			System.exit( 0 );

		}

	}


/*	class RemindTask extends TimerTask {
		public void run() {
			//Send packet to the mote to get next line
			RWFlashMsg RWFlashPkt = new RWFlashMsg();
			RWFlashPkt.set_top(QueryController.READ_LINE);
			RWFlashPkt.set_line_no(read_line_counter);
			RWFlashPkt.set_seq_no(TOTAL_LINE_COUNT);
			RWFlashPkt.set_dest(dest_node);
			try {
				mote.send(dest_node, RWFlashPkt);
				currLinesRead = 0;
				System.out.println("Sent out REPEAT request starting from line no " + read_line_counter);
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
*/

 }
